/*
Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Amsterdam nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * File: 2DShapes.cpp
 *
 * Description: Implementation of 2D Shape Classes
 *
 * Author: 3D Robocup Soccer Team 83 <persisgroup@aim.com>, (c) 2006
 *
 */

#include <algorithm>
#include "2DShapes.h"
#include "PolarVector.h"

using namespace std;

/*****************************************************************************/
/********************* CLASS CIRCLE ******************************************/
/*****************************************************************************/

/*! This is the constructor of a circle.
    \param pos first point that defines the center of circle
    \param dR the radius of the circle
    \return circle with pos as center and radius as radius*/
Circle::Circle( Vector3D pos, double dR )
{
    setCircle( pos, dR );
}

/*! This is the constructor of a circle which initializes a circle with a
    radius of zero. */
Circle::Circle( )
{
    setCircle( Vector3D( -1000.0, -1000.0 ), 0 );
}

/*! This method prints the circle information to the specified output stream
    in the following format: "c: (c_x,c_y), r: rad" where (c_x,c_y) denotes
    the center of the circle and rad the radius.
    \param os output stream to which output is written. */
ostream& operator << ( std::ostream &os, Circle c )
{
    os << "c:" << c.m_posCenter << ", r:" << c.m_dRadius;
    return os;
}

/*! This method sets the values of the circle.
    \param pos new center of the circle
    \param dR new radius of the circle
    ( > 0 )
    \return bool indicating whether radius was set */
bool Circle::setCircle( Vector3D pos, double dR )
{
    setCenter( pos );
    return setRadius( dR  );
}
/*! This method sets the radius of the circle.
    \param dR new radius of the circle ( > 0 )
    \return bool indicating whether radius was set */
bool Circle::setRadius( double dR )
{
    if( dR > 0 )
    {
        m_dRadius = dR;
        return true;
    }
    else
    {
        m_dRadius = 0.0;
        return false;
    }
}

/*! This method returns the radius of the circle.
    \return radius of the circle */
double Circle::getRadius()
{
    return m_dRadius;
}

/*! This method sets the center of the circle.
    \param pos new center of the circle
    \return bool indicating whether center was set */
bool Circle::setCenter( Vector3D pos )
{
    m_posCenter = pos;
    return true;
}

/*! This method returns the center of the circle.
    \return center of the circle */
Vector3D Circle::getCenter()
{
    return m_posCenter;
}

/*! This method returns the circumference of the circle.
    \return circumference of the circle */
double Circle::getCircumference()
{
    return 2.0 * M_PI * getRadius();
}

/*! This method returns the area inside the circle.
    \return area inside the circle */
double Circle::getArea()
{
    return M_PI * getRadius() * getRadius();
}

/*! This method returns a boolean that indicates whether 'pos' is
    located inside the circle.

   \param pos position of which should be checked whether it is
   located in the circle

   \return bool indicating whether pos lies inside the circle */
bool Circle::isInside( Vector3D pos )
{
    return m_posCenter.GetDistanceTo( pos ) < getRadius() ;
}

/*! This method returns the two possible intersection points between two
    circles. This method returns the number of solutions that were found.
    \param c circle with which intersection should be found
    \param p1 will be filled with first solution
    \param p2 will be filled with second solution
    \return number of solutions. */
int Circle::getIntersectionPoints( Circle c, Vector3D *p1, Vector3D *p2 )
{
    double x0, y0, r0;
    double x1, y1, r1;

    x0 = getCenter( ).x();
    y0 = getCenter( ).y();
    r0 = getRadius( );
    x1 = c.getCenter( ).x();
    y1 = c.getCenter( ).y();
    r1 = c.getRadius( );

    double      d, dx, dy, h, a, x, y, p2_x, p2_y;

    // first calculate distance between two centers circles P0 and P1.
    dx = x1 - x0;
    dy = y1 - y0;
    d = sqrt( dx * dx + dy * dy );

    // normalize differences
    dx /= d; dy /= d;

    // a is distance between p0 and point that is the intersection point P2
    // that intersects P0-P1 and the line that crosses the two intersection
    // points P3 and P4.
    // Define two triangles: P0,P2,P3 and P1,P2,P3.
    // with distances a, h, r0 and b, h, r1 with d = a + b
    // We know a^2 + h^2 = r0^2 and b^2 + h^2 = r1^2 which then gives
    // a^2 + r1^2 - b^2 = r0^2 with d = a + b ==> a^2 + r1^2 - (d-a)^2 = r0^2
    // ==> r0^2 + d^2 - r1^2 / 2*d
    a = ( r0 * r0 + d * d - r1 * r1 ) / ( 2.0 * d );

    // h is then a^2 + h^2 = r0^2 ==> h = sqrt( r0^2 - a^2 )
    double      arg = r0 * r0 - a * a;
    h = ( arg > 0.0 ) ? sqrt( arg ) : 0.0;

    // First calculate P2
    p2_x = x0 + a * dx;
    p2_y = y0 + a * dy;

    // and finally the two intersection points
    x =  p2_x - h * dy;
    y =  p2_y + h * dx;
    p1->Set( x, y );
    x =  p2_x + h * dy;
    y =  p2_y - h * dx;
    p2->Set( x, y );

    return ( arg < 0.0 ) ? 0 : ( ( arg == 0.0 ) ? 1 :  2 );
}

/*! This method returns the size of the intersection area of two circles.
    \param c circle with which intersection should be determined
    \return size of the intersection area. */
double Circle::getIntersectionArea( Circle c )
{
    Vector3D pos1, pos2, pos3;
    double d, h, dArea;
    AngDeg ang;

    d = getCenter().GetDistanceTo( c.getCenter() ); // dist between two centers
    if( d > c.getRadius() + getRadius() )           // larger than sum radii
        return 0.0;                                   // circles do not intersect
    if( d <= fabs( c.getRadius() - getRadius() ) )   // one totally in the other
    {
        double dR = min( c.getRadius(), getRadius() );// return area smallest circ
        return M_PI*dR*dR;
    }

    int iNrSol = getIntersectionPoints( c, &pos1, &pos2 );
    if( iNrSol != 2 )
        return 0.0;

    // the intersection area of two circles can be divided into two segments:
    // left and right of the line between the two intersection points p1 and p2.
    // The outside area of each segment can be calculated by taking the part
    // of the circle pie excluding the triangle from the center to the
    // two intersection points.
    // The pie equals pi*r^2 * rad(2*ang) / 2*pi = 0.5*rad(2*ang)*r^2 with ang
    // the angle between the center c of the circle and one of the two
    // intersection points. Thus the angle between c and p1 and c and p3 where
    // p3 is the point that lies halfway between p1 and p2.
    // This can be calculated using ang = asin( d / r ) with d the distance
    // between p1 and p3 and r the radius of the circle.
    // The area of the triangle is 2*0.5*h*d.

    pos3 = ( pos2 - pos1 ).LinearInterpolate( Vector3D(0,0,0), 0.5 );
    d = pos1.GetDistanceTo( pos3 );
    h = pos3.GetDistanceTo( getCenter() );
    ang = asin( d / getRadius() );

    dArea = ang * getRadius() * getRadius();
    dArea = dArea - d * h;

    // and now for the other segment the same story
    h = pos3.GetDistanceTo( c.getCenter() );
    ang = asin( d / c.getRadius() );
    dArea = dArea + ang * c.getRadius() * c.getRadius();
    dArea = dArea - d * h;

    return dArea;
}


/*****************************************************************************/
/***********************  CLASS LINE *****************************************/
/*****************************************************************************/

/*! This constructor creates a line by given the three coefficents of the line.
    A line is specified by the formula ay + bx + c = 0.
    \param dA a coefficients of the line
    \param dB b coefficients of the line
    \param dC c coefficients of the line */
Line::Line( double dA, double dB, double dC )
{
    m_a = dA;
    m_b = dB;
    m_c = dC;
}

/*! This function prints the line to the specified output stream in the
    format y = ax + b.
    \param os output stream to which output is written
    \param l line that is written to output stream
    \return output sream to which output is appended. */
ostream& operator <<( ostream & os, Line l )
{
    double a = l.getACoefficient();
    double b = l.getBCoefficient();
    double c = l.getCCoefficient();

    // ay + bx + c = 0 -> y = -b/a x - c/a
    if( a == 0 )
        os << "x = " << -c / b;
    else
    {
        os << "y = ";
        if( b != 0 )
            os << -b / a << "x ";
        if( c > 0 )
            os << "- " <<  fabs( c / a );
        else if( c < 0 )
            os << "+ " <<  fabs( c / a );
    }
    return os;
}

/*! This method returns the intersection point between the current Line and
    the specified line.
    \param line line with which the intersection should be calculated.
    \return Vector3D position that is the intersection point. */
Vector3D Line::getIntersection( Line line )
{
    Vector3D pos;
    double x, y;
    if( ( m_a / m_b ) ==  ( line.getACoefficient() / line.getBCoefficient() ) )
        return pos; // lines are parallel, no intersection
    if( m_a == 0 )            // bx + c = 0 and a2*y + b2*x + c2 = 0 ==> x = -c/b
    {                          // calculate x using the current line
        x = -m_c / m_b;                // and calculate the y using the second line
        y = line.getYGivenX( x );
    }
    else if( line.getACoefficient() == 0 )
    {                         // ay + bx + c = 0 and b2*x + c2 = 0 ==> x = -c2/b2
        x = -line.getCCoefficient() / line.getBCoefficient(); // calculate x using
        y = getYGivenX( x );       // 2nd line and calculate y using current line
    }
    // ay + bx + c = 0 and a2y + b2*x + c2 = 0
    // y = (-b2/a2)x - c2/a2
    // bx = -a*y - c =>  bx = -a*(-b2/a2)x -a*(-c2/a2) - c ==>
    // ==> a2*bx = a*b2*x + a*c2 - a2*c ==> x = (a*c2 - a2*c)/(a2*b - a*b2)
    // calculate x using the above formula and the y using the current line
    else
    {
        x = ( m_a * line.getCCoefficient() - line.getACoefficient() * m_c ) /
            ( line.getACoefficient() * m_b - m_a * line.getBCoefficient() );
        y = getYGivenX( x );
    }

    return Vector3D( x, y );
}


/*! This method calculates the intersection points between the current line
    and the circle specified with as center 'posCenter' and radius 'dRadius'.
    The number of solutions are returned and the corresponding points are put
    in the third and fourth argument of the method
    \param c circle with which intersection points should be found
    \param posSolution1 first intersection (if any)
    \param posSolution2 second intersection (if any) */
int Line::getCircleIntersectionPoints( Circle circle,
                                       Vector3D *posSolution1, Vector3D *posSolution2 )
{
    int    iSol;
    double dSol1 = 0, dSol2 = 0;
    double h = circle.getCenter().x();
    double k = circle.getCenter().y();

    // line:   x = -c/b (if a = 0)
    // circle: (x-h)^2 + (y-k)^2 = r^2, with h = center.x and k = center.y
    // fill in:(-c/b-h)^2 + y^2 -2ky + k^2 - r^2 = 0
    //         y^2 -2ky + (-c/b-h)^2 + k^2 - r^2 = 0
    // and determine solutions for y using abc-formula
    if( fabs( m_a ) < EPSILON )
    {
        iSol = Geometry::abcFormula( 1, -2 * k, ( ( -m_c / m_b ) - h ) * ( ( -m_c / m_b ) - h )
                                     + k * k - circle.getRadius() * circle.getRadius(), &dSol1, &dSol2 );
        posSolution1->Set( ( -m_c / m_b ), dSol1 );
        posSolution2->Set( ( -m_c / m_b ), dSol2 );
        return iSol;
    }

    // ay + bx + c = 0 => y = -b/a x - c/a, with da = -b/a and db = -c/a
    // circle: (x-h)^2 + (y-k)^2 = r^2, with h = center.x and k = center.y
    // fill in:x^2 -2hx + h^2 + (da*x-db)^2 -2k(da*x-db) + k^2 - r^2 = 0
    //         x^2 -2hx + h^2 + da^2*x^2 + 2da*db*x + db^2 -2k*da*x -2k*db
    //                                                         + k^2 - r^2 = 0
    //       (1+da^2)*x^2 + 2(da*db-h-k*da)*x + h2 + db^2  -2k*db + k^2 - r^2 = 0
    // and determine solutions for x using abc-formula
    // fill in x in original line equation to get y coordinate
    double da = -m_b / m_a;
    double db = -m_c / m_a;

    double dA = 1 + da * da;
    double dB = 2 * ( da * db - h - k * da );
    double dC = h * h + db * db - 2 * k * db + k * k - circle.getRadius() * circle.getRadius();

    iSol = Geometry::abcFormula( dA, dB, dC, &dSol1, &dSol2 );

    posSolution1->Set( dSol1, da * dSol1 + db );
    posSolution2->Set( dSol2, da * dSol2 + db );
    return iSol;

}

/*! This method returns the tangent line to a Vector3D. This is the line
    between the specified position and the closest point on the line to this
    position.
    \param pos Vector3D point with which tangent line is calculated.
    \return Line line tangent to this position */
Line Line::getTangentLine( Vector3D pos )
{
    // ay + bx + c = 0 -> y = (-b/a)x + (-c/a)
    // tangent: y = (a/b)*x + C1 -> by - ax + C2 = 0 => C2 = ax - by
    // with pos.y = y, pos.x = x
    return Line( m_b, -m_a, m_a * pos.x() - m_b * pos.y() );
}

/*! This method returns the closest point on a line to a given position.
    \param pos point to which closest point should be determined
    \return Vector3D closest point on line to 'pos'. */
Vector3D Line::getPointOnLineClosestTo( Vector3D pos )
{
    Line l2 = getTangentLine( pos );  // get tangent line
    return getIntersection( l2 );     // and intersection between the two lines
}

/*! This method returns the distance between a specified position and the
    closest point on the given line.
    \param pos position to which distance should be calculated
    \return double indicating the distance to the line. */
double Line::getDistanceWithPoint( Vector3D pos )
{
    return pos.GetDistanceTo( getPointOnLineClosestTo( pos ) );
}

/*! This method determines whether the projection of a point on the
    current line lies between two other points ('point1' and 'point2')
    that lie on the same line.

    \param pos point of which projection is checked.
    \param point1 first point on line
    \param point2 second point on line
    \return true when projection of 'pos' lies between 'point1' and 'point2'.*/
bool Line::isInBetween( Vector3D pos, Vector3D point1, Vector3D point2 )
{
    pos          = getPointOnLineClosestTo( pos ); // get closest point
    double dDist = point1.GetDistanceTo( point2 ); // get distance between 2 pos

    // if the distance from both points to the projection is smaller than this
    // dist, the pos lies in between.
    return pos.GetDistanceTo( point1 ) <= dDist &&
           pos.GetDistanceTo( point2 ) <= dDist;
}

/*! This method calculates the y coordinate given the x coordinate
    \param x coordinate
    \return y coordinate on this line */
double Line::getYGivenX( double x )
{
    if( m_a == 0 )
    {
        cerr << "(Line::yGivenX) Cannot calculate Y coordinate: " ;
//   show( cerr );
        cerr << endl;
        return 0;
    }
    // ay + bx + c = 0 ==> ay = -(b*x + c)/a
    return -( m_b * x + m_c ) / m_a;
}

/*! This method calculates the x coordinate given the x coordinate
    \param y coordinate
    \return x coordinate on this line */
double Line::getXGivenY( double y )
{
    if( m_b == 0 )
    {
        cerr << "(Line::xGivenY) Cannot calculate X coordinate\n" ;
        return 0;
    }
    // ay + bx + c = 0 ==> bx = -(a*y + c)/a
    return -( m_a * y + m_c ) / m_b;
}

/*! This method creates a line given two points.
    \param pos1 first point
    \param pos2 second point
    \return line that passes through the two specified points. */
Line Line::makeLineFromTwoPoints( Vector3D pos1, Vector3D pos2 )
{
    // 1*y + bx + c = 0 => y = -bx - c
    // with -b the direction coefficient (or slope)
    // and c = - y - bx
    double dA, dB, dC;
    double dTemp = pos2.x() - pos1.x(); // determine the slope
    if( fabs( dTemp ) < EPSILON )
    {
        // ay + bx + c = 0 with vertical slope=> a = 0, b = 1
        dA = 0.0;
        dB = 1.0;
    }
    else
    {
        // y = (-b)x -c with -b the slope of the line
        dA = 1.0;
        dB = -( pos2.y() - pos1.y() ) / dTemp;
    }
    // ay + bx + c = 0 ==> c = -a*y - b*x
    dC =  - dA * pos2.y()  - dB * pos2.x();
    return Line( dA, dB, dC );
}

/*! This method creates a line given a position and an angle.
    \param vec position through which the line passes
    \param angle direction of the line.
    \return line that goes through position 'vec' with angle 'angle'. */
Line Line::makeLineFromPositionAndAngle( Vector3D vec, AngDeg angle )
{
    // calculate point somewhat further in direction 'angle' and make
    // line from these two points.
    return makeLineFromTwoPoints( vec, vec + PolarVector3D( 1, angle ) );
}

/*! This method returns the a coefficient from the line ay + bx + c = 0.
    \return a coefficient of the line. */
double Line::getACoefficient() const
{
    return m_a;
}

/*! This method returns the b coefficient from the line ay + bx + c = 0.
    \return b coefficient of the line. */
double Line::getBCoefficient() const
{
    return m_b;
}

/*! This method returns the c coefficient from the line ay + bx + c = 0.
    \return c coefficient of the line. */
double Line::getCCoefficient() const
{
    return m_c;
}

/*****************************************************************************/
/********************* CLASS RECTANGLE ***************************************/
/*****************************************************************************/

/*! This is the constructor of a Rectangle. Two points will be given. The
    order does not matter as long as two opposite points are given (left
    top and right bottom or right top and left bottom).
    \param pos first point that defines corner of rectangle
    \param pos2 second point that defines other corner of rectangle
    \return rectangle with 'pos' and 'pos2' as opposite corners. */
Rect::Rect( Vector3D pos, Vector3D pos2 )
{
    setRectanglePoints( pos, pos2 );
}

/*! This method sets the upper left and right bottom point of the current
    rectangle.
    \param pos first point that defines corner of rectangle
    \param pos2 second point that defines other corner of rectangle */
void Rect::setRectanglePoints( Vector3D pos1, Vector3D pos2 )
{
    m_posLeftTop.x() = max( pos1.x(), pos2.x() );
    m_posLeftTop.y() = min( pos1.y(), pos2.y() );
    m_posRightBottom.x() = min( pos1.x(), pos2.x() );
    m_posRightBottom.y() = max( pos1.y(), pos2.y() );
}

/*! This method prints the rectangle to the specified output stream in the
    format rect( top_left_point, bottom_right_point ).
    \param os output stream to which rectangle is written. */
ostream& operator<<( ostream &os, Rect r )
{
    os << "rect(" << r.m_posLeftTop << " " << r.m_posRightBottom << ")";
    return os;
}

/*! This method determines whether the given position lies inside the current
    rectangle.
    \param pos position which is checked whether it lies in rectangle
    \return true when 'pos' lies in the rectangle, false otherwise */
bool Rect::isInside( Vector3D pos )
{
    return pos.IsBetweenX( m_posRightBottom.x(), m_posLeftTop.x() ) &&
           pos.IsBetweenY( m_posLeftTop.y(),     m_posRightBottom.y() );

}

/*! This method sets the top left position of the rectangle
    \param pos new top left position of the rectangle
    \return true when update was successful */
bool Rect::setPosLeftTop( Vector3D pos )
{
    m_posLeftTop = pos;
    return true;
}

/*! This method returns the top left position of the rectangle
    \return top left position of the rectangle */
Vector3D Rect::getPosLeftTop(  )
{
    return m_posLeftTop;
}

/*! This method sets the right bottom position of the rectangle
    \param pos new right bottom position of the rectangle
    \return true when update was succesfull */
bool Rect::setPosRightBottom( Vector3D pos )
{
    m_posRightBottom = pos;
    return true;
}

/*! This method returns the right bottom position of the rectangle
    \return top right bottom of the rectangle */
Vector3D Rect::getPosRightBottom(  )
{
    return m_posRightBottom;
}
