/*
Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

3. Neither the name of the University of Amsterdam nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * File: 2DShapes.h
 *
 * Description: 2D Shape Classes
 *
 * Author: 3D Robocup Soccer Team 83 <persisgroup@aim.com>, (c) 2006
 *
 */

#ifndef _2DSHAPES_H_
#define _2DSHAPES_H_

#include "MathUtils.h"
#include "Vector.h"

/*****************************************************************************/
/********************** CLASS CIRCLE *****************************************/
/*****************************************************************************/

/*!This class represents a circle. A circle is defined by one Vector2f
 (which denotes the center) and its radius. */
class Circle
{
        Vector3D m_posCenter; /*!< Center of the circle  */
        double m_dRadius; /*!< Radius of the circle  */

    public:
        Circle();
        Circle(Vector3D pos, double dR);

        friend std::ostream& operator <<(std::ostream &os, Circle c);

        // get and set methods
        bool setCircle(Vector3D pos, double dR);
        bool setRadius(double dR);
        double getRadius();
        bool setCenter(Vector3D pos);
        Vector3D getCenter();
        double getCircumference();
        double getArea();

        // calculate intersection points and area with other circle
        bool isInside(Vector3D pos);
        int getIntersectionPoints(Circle c, Vector3D *p1, Vector3D *p2);
        double getIntersectionArea(Circle c);

};

/*****************************************************************************/
/*********************** CLASS LINE ******************************************/
/*****************************************************************************/

/*!This class contains the representation of a line. A line is defined
 by the formula ay + bx + c = 0. The coefficients a, b and c are stored
 and used in the calculations. */
class Line
{
        // a line is defined by the formula: ay + bx + c = 0
        double m_a; /*!< This is the a coefficient in the line ay + bx + c = 0 */
        double m_b; /*!< This is the b coefficient in the line ay + bx + c = 0 */
        double m_c; /*!< This is the c coefficient in the line ay + bx + c = 0 */

    public:
        Line(double a, double b, double c);

        // print method
        friend std::ostream& operator <<(std::ostream & os, Line l);

        // get intersection points with this line
        Vector3D getIntersection(Line line);
        int getCircleIntersectionPoints(Circle circle, Vector3D *posSolution1,
                Vector3D *posSolution2);
        Line getTangentLine(Vector3D pos);
        Vector3D getPointOnLineClosestTo(Vector3D pos);
        double getDistanceWithPoint(Vector3D pos);
        bool isInBetween(Vector3D pos, Vector3D point1, Vector3D point2);

        // calculate associated variables in the line
        double getYGivenX(double x);
        double getXGivenY(double y);
        double getACoefficient() const;
        double getBCoefficient() const;
        double getCCoefficient() const;

        // static methods to make a line using an easier representation.
        static Line makeLineFromTwoPoints(Vector3D pos1, Vector3D pos2);
        static Line makeLineFromPositionAndAngle(Vector3D vec, AngDeg angle);
};

/*****************************************************************************/
/********************** CLASS RECTANGLE **************************************/
/******************************************************************************/
/*!This class represents a rectangle. A rectangle is defined by two
 Vector2Ds the one at the upper left corner and the one at the
 right bottom. */
class Rect
{
        Vector3D m_posLeftTop; /*!< top left position of the rectangle      */
        Vector3D m_posRightBottom; /*!< bottom right position of the rectangle  */

    public:
        Rect(Vector3D pos, Vector3D pos2);

        friend std::ostream& operator <<(std::ostream &os, Rect t);

        // checks whether point lies inside the rectangle
        bool isInside(Vector3D pos);

        // standard get and set methosd
        void setRectanglePoints(Vector3D pos1, Vector3D pos2);
        bool setPosLeftTop(Vector3D pos);
        Vector3D getPosLeftTop();
        bool setPosRightBottom(Vector3D pos);
        Vector3D getPosRightBottom();
};

#endif /*2DSHAPES_H_*/
