/*
 * File: ActionManager.cpp
 *
 * Description: ActionManager class implementations
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ActionManager.h"
#include "LogLevels.h"
#include "ParseTree.h"
#include "AgentBodyStates.h"
#include "WorldModel.h"
using namespace std;

ActionManager::ActionManager(WorldModel &wm) :
    transitions(agentStates), currentActType(ILLEGAL_ACT), nextActType(ILLEGAL_ACT),
            stepStartTime(0), world(wm),
            prevState(AgentBodyStates::ILLEGAL_STATE),
            nextState(AgentBodyStates::ILLEGAL_STATE),
            currentAction(NULL),
            currentTransition(NULL), stateChanged(false)
{
}

ActionManager::~ActionManager()
{
}

void ActionManager::SetActionType( int acttype, const char *name )
{
    if (static_cast<unsigned>(acttype) > actionIDs.size())
    {
        Log(LOG_ERROR) << "Invalid act type requested!";
        return;
    }
    map<string, int>::const_iterator numptr = actionNumberMap.find(name);
    if (numptr == actionNumberMap.end())
        Log(LOG_ERROR) << "Action not found: " << name;
    else
        actionIDs[acttype] = numptr->second;
}

void ActionManager::LoadFile(const char *filename)
{
    ParseTree parser(ParseTree::LoadFileToString(filename));
    for (unsigned i = 0; i < parser.GetNumOfKeys(); ++i) {
        if (parser.GetKey(i) == "State")
            agentStates.LoadState(parser.GetData(i));
        else if (parser.GetKey(i) == "Transition")
            transitions.LoadTransition(parser.GetData(i));
        else if (parser.GetKey(i) == "Action")
        {
            ParseTree *act = parser.GetData(i);
            if (act->GetKey(0)!= "name")
            {
                Log(LOG_ACTION_MANAGER) << "Loading failed, no name field found!";
                return;
            }
            string name = act->GetData(0)->GetKey(0);
            Log(LOG_ACTION_MANAGER) << "\nLoading action: " << name;
            actionNumberMap[name] = actions.size();
            Action tmp;
            tmp.LoadAction(act, agentStates);
            actions.push_back(tmp);
        }
        else
            Log(LOG_ERROR) << "Action FileLoader: unknown tag encountered -> "
                    << parser.GetKey(i);
    }

    actionIDs.resize(actions.size());
}

int ActionManager::ActionType(const char *actName) const
{
    int type = ILLEGAL_ACT;
    map<string, int>::const_iterator numptr = actionNumberMap.find(actName);
    if (numptr == actionNumberMap.end())
        Log(LOG_ERROR) << "Action not found: " << actName;
    else
    {
        for (unsigned i = 0; i < actionIDs.size(); ++i)
        {
            if (actionIDs[i] == numptr->second)
                type = i;
        }
    }
    return type;
}

void ActionManager::GotoNextStep()
{
    stateChanged = false;
    stepStartTime = world.Time();
    assert(currentTransition!=NULL);
    if (!currentTransition->GotoNextStep())
    {
        stateChanged = true;
        currentTransition->Reset();
        do
        {
            currentAction->ChangeState();
            prevState = nextState;
            nextState = currentAction->NextState();

            if (nextState == AgentBodyStates::ILLEGAL_STATE)
            {
                Log(LOG_ACTION_MANAGER) << "Switching to the next action..."
                        << nextActType;
                currentAction->Reset();
                currentActType = nextActType;
                nextActType = ILLEGAL_ACT;
                if (currentActType == ILLEGAL_ACT)
                {
                    currentAction = NULL;
                    nextState = prevState = AgentBodyStates::ILLEGAL_STATE;
                    return;
                }
                currentAction = &actions[actionIDs[currentActType]];
                nextState = currentAction->FirstState();
                currentAction->Lock();
            }

            currentTransition = transitions.Transition(prevState, nextState);
        } while (!currentTransition);
        Log(LOG_ACTION_MANAGER) << "New step " << currentTransition->StepName();
        Log(LOG_ACTION_MANAGER) << "Changed state to "<< agentStates.StateName(prevState)
                << " going to "<< agentStates.StateName(nextState);
    }
    else
        Log(LOG_ACTION_MANAGER) << "Acting in step " << currentTransition->StepName();
}

void ActionManager::StartActing(int act)
{
    stateChanged = true;
    stepStartTime = world.Time();
    currentActType = act;
    currentAction = &actions[actionIDs[currentActType]];
    prevState = currentAction->FirstState();
    nextState = currentAction->NextState();
    currentTransition = transitions.Transition(prevState, nextState);
    while (currentTransition == 0)
    {
        currentAction->ChangeState();
        prevState = nextState;
        nextState = currentAction->NextState();

        if (nextState == AgentBodyStates::ILLEGAL_STATE)
        {
            Log(LOG_ACTION_MANAGER) << "Can't start action!";
            return;
        }

        currentTransition = transitions.Transition(prevState, nextState);
    }
    Log(LOG_ACTION_MANAGER) << "Start act "<< currentActType << " in step: "
            << currentTransition->StepName();
    Log(LOG_ACTION_MANAGER) << "From state "<< agentStates.StateName(prevState) << " to "
            << agentStates.StateName(nextState);
}
