/*
 * File: MathUtils.h
 *
 * Description: Mathematical utilities
 *
 * Author: 3D Robocup Soccer Team 83 <persisgroup@aim.com>, (c) 2006
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef MATHUTILS_H_
#define MATHUTILS_H_

#include <salt/gmath.h>
#include <cmath>
#include <limits>

using namespace salt;

#define EPSILON 0.0001

typedef float AngDeg;
typedef double AngRad;

extern const AngDeg UnknownAngle;
extern const AngRad UnknownRadian;
extern const double UnknownDouble;
extern const float  UnknownFloat;
extern const int    UnknownInteger;

inline int sign( double n )
{
    return n >= 0 ? 1 : -1;
}

template<class T>
inline T sqr( T x )
{
    return x*x;
}

inline double cosDeg( AngDeg x )
{
    return gCos(gDegToRad(x));
}

inline double sinDeg( AngDeg x )
{
    return gSin(gDegToRad(x));
}

inline double tanDeg( AngDeg x )
{
    return tan(gDegToRad(x));
}

inline AngDeg atanDeg( double x )
{
    return gRadToDeg(gArcTan(x));
}

inline double atan2Deg( double x, double y )
{
    return gRadToDeg(gArcTan2(x,y));
}

inline AngDeg acosDeg( double x )
{
    return gRadToDeg(gArcCos(x));
}

inline AngDeg asinDeg( double x )
{
    return gRadToDeg(gArcSin(x));
}

inline double GetNProb( double err, double sig )
{
    return 1.0 / (sig * sqrt( 2 * M_PI )) * exp( -sqr( err ) / ( 2 * sqr( sig ) ) );
}

inline float fabs_max(float x, float y)
{
    return fabs(x)>fabs(y) ? true : false;
}


/*****************************************************************************/
/*********************   CLASS GEOMETRY   ************************************/
/*****************************************************************************/

/*! This class contains several static methods dealing with geometry.*/
class Geometry
{

    public:

        // geometric series
        static double getLengthGeomSeries( double dFirst, double dRatio, double dSum  );
        static double getSumGeomSeries   ( double dFirst, double dRatio, double dLen  );
        static double getSumInfGeomSeries( double dFirst, double dRatio              );
        static double getFirstGeomSeries ( double dSum,  double dRatio, double dLen  );
        static double getFirstInfGeomSeries( double dSum, double dRatio              );

        // abc formula
        static int    abcFormula( double a, double b, double c, double *s1, double *s2 );
};

#endif /*MATHUTILS_H_*/
