/*
 * File: TransitionData.h
 *
 * Description: State Transition Data Class
 *
 * Author: UI-AI 3D Robocup Soccer Team <ui_3dsoccer@eng.ui.ac.ir>, (c) 2007
 *
 */

/*
    Copyright (C) 2009  UI-AI 3D Team

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef TRANSITIONDATA_H_
#define TRANSITIONDATA_H_

#include <vector>
#include <string>
#include "LogLevels.h"
#include "AgentBody.h"

using namespace std;

/**
 * This class contains the complete data about a single transition.
 */
class TransitionData
{
    public:
        /**
         * This class contains the data of a single joint action. Maybe merged with
         * the AgentBody::JointInfo into a more general class.
         */
        class JointTransitionData
        {
            public:
                AgentBody::Joint joint;
                union {
                    float rate[2];
                    float delay;
                };
                float targetax[2];

                JointTransitionData() : joint(AgentBody::ILLEGAL_JOINT)
                {
                    rate[0] = rate[1] = UnknownAngle;
                    targetax[0] = targetax[1] = UnknownAngle;
                }

                void ReadJointData(ParseTree *jointInfo);
                void UpdateValues(const TransitionData *actions);

            private:
                string rateparam[2];
                string targetaxparam[2];

                void ReadParam(string str, float &val, string &param);
        };

        typedef vector<JointTransitionData> JointTransitionVector;

        struct StepData
        {
            StepData(const string &name, bool setAll = false ): setAllJoints(setAll), name(name)
            {
            }

            bool setAllJoints;  //< Indicates if we should set rate for all of the joints
            string name;
            string checkFunction;
            JointTransitionVector actionList;
        };

    public:
        TransitionData(): currentStep(0)
        {
        }

        JointTransitionVector &JointActions()	{ return steps[currentStep].actionList; }
        void Reset()							{ currentStep = 0; }
        int CurrentStep() const					{ return currentStep; }
        const string &StepName() const			{ return steps[currentStep].name; }
        const string &CheckFunction() const     { return steps[currentStep].checkFunction; }
        void SetParameter(string name, float value) { parameters[name] = value; }
        bool IsTerminated() const				{ return currentStep >= steps.size(); }

        float GetParameter(string name) const;
        void ReadTransitionData( ParseTree *transition );

        void UpdateParameters()
        {
            for (unsigned i = 0; i < steps[currentStep].actionList.size(); ++i)
                steps[currentStep].actionList[i].UpdateValues(this);
        }

        bool GotoNextStep()
        {
            currentStep++;
            return !IsTerminated();
        }

    private:
        vector<StepData> steps;
        unsigned currentStep;
        map<string, float> parameters;

        void AddJointAction(const JointTransitionData &jointData);
};

inline float TransitionData::GetParameter(string name) const
{
    map<string, float>::const_iterator it = parameters.find(name);
    if (it==parameters.end())
    {
        Log(LOG_WARNING) << "Invalid parameter: " << name << " requested!";
        return 0.0;
    }
    else
        return it->second;
}

#endif /*TRANSITIONDATA_H_*/
